import metallurgy as mg


def test_alloy_creation():

    alloy_from_string = mg.Alloy("Fe20Ag80")

    alloy_from_dict = mg.Alloy({"Fe": 20, "Ag": 80})

    assert alloy_from_string.__dict__ == alloy_from_dict.__dict__


def test_alloy_to_string():
    alloy = mg.Alloy({"Fe": 20, "Ag": 80})
    assert alloy.to_string() == "Ag80Fe20"


def test_alloy_element_deletion():

    num_alloys = 100
    random_alloys = mg.generate.random_alloys(
        num_alloys,
        min_elements=2,
        max_elements=2,
    )

    for alloy in random_alloys:
        alloy.remove_element(alloy.elements[0])
        assert alloy.to_string() == alloy.elements[0] + "100"

    random_alloys = mg.generate.random_alloys(
        num_alloys,
        min_elements=2,
        max_elements=2,
    )

    for alloy in random_alloys:
        del alloy.composition[alloy.elements[0]]
        assert alloy.to_string() == alloy.elements[0] + "100"



"""Module enabling calculation of deviations of elemental properties in alloys."""


from typing import Union, List
from numbers import Number
from collections.abc import Iterable

import numpy as np
import metallurgy as mg


def deviation(
    alloy: Union[mg.Alloy, str, dict], property_name: str
) -> Union[float, None, List[Union[float, None]]]:
    """Returns the deviation of a particular elemental property in an
    alloy.

    See equation 6 of the paper "Machine-learning improves understanding of
    glass formation in metallic systems" for definition of the deviation:
    https://pubs.rsc.org/en/content/articlelanding/2022/dd/d2dd00026a

    :group: calculations

    Parameters
    ----------

    alloy
        The alloy for which to calculate the linear mixture.
    property_name
        The elemental property to calcualate the linear mixture of.

    """

    # If a list of alloys is given, return a list of deviation data
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [deviation(a, property_name) for a in list(alloy)]

    # Convert input alloy to an Alloy instance if not already
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    # Deviation only makes sense for multi-element alloys
    if len(alloy.elements) > 1:

        # If property is numerical, calculate the deviation of the values
        if isinstance(
            mg.periodic_table.elements[alloy.elements[0]][property_name],
            (Number, list),
        ):

            # Calculate the mean value of the property in the alloy
            mean = 0
            for element in alloy.elements:
                value = mg.periodic_table.elements[element][property_name]

                # Return None if property is None
                if value is None:
                    return None

                # Take first entry if property is a list
                if isinstance(value, list):
                    value = value[0]

                # Return None if property is not numerical
                if not isinstance(value, Number):
                    return None

                mean += alloy.composition[element] * value

            # Calculate the deviation of the property in the alloy
            total_deviation = 0
            for element in alloy.elements:
                value = mg.periodic_table.elements[element][property_name]

                # Return None if property is None
                if value is None:
                    return None

                # Take first entry if property is a list
                if isinstance(value, list):
                    value = value[0]

                # Return None if property is not numerical
                if not isinstance(value, Number):
                    return None

                total_deviation += alloy.composition[element] * (
                    (value - mean) ** 2
                )

            return total_deviation**0.5

        # If property is non-numerical, calculate the shannon entropy of the
        # values
        else:

            # Count unique values of the property
            value_count = {}
            for element in alloy.elements:
                value = mg.periodic_table.elements[element][property_name]
                if value is None:
                    return None

                if value not in value_count:
                    value_count[value] = 0
                value_count[value] += alloy.composition[element]

            # Shannon entropy is only non-zero if multiple unique values
            if len(value_count) > 1:
                shannonEntropy = 0
                for value in value_count:
                    shannonEntropy -= value_count[value] * np.log(
                        value_count[value]
                    )
                return shannonEntropy
            else:
                return 0.0

    # Single element alloys have zero deviation
    else:
        return 0.0



"""Module providing an interface to all other calculation functions provided by
metallurgy.
"""


import inspect
from dataclasses import fields
from typing import Callable, List, Optional, Union

import elementy

import metallurgy as mg


def is_mod_function(mod, func):
    """checks that func is a function defined in module mod"""
    return inspect.isfunction(func) and inspect.getmodule(func) == mod


def list_functions(mod):
    """list of functions defined in module mod"""
    return [
        func.__name__
        for func in mod.__dict__.values()
        if is_mod_function(mod, func)
    ]


def get_property_function(property_name: str) -> Union[Callable, None]:
    """Get the function responsible for calculating a particular property, if it
    exists in metallurgy.

    :group: utils

    Parameters
    ----------

    property_name
        Name of the property to get the function for.

    """

    # Get all modules in metallurgy
    modules = inspect.getmembers(mg, inspect.ismodule)
    for module in modules:
        # For each function in the module, check if the name matches the property
        for func in list_functions(module[1]):
            if func == property_name:
                return getattr(module[1], func)


def get_all_properties() -> List[str]:
    """Returns every calculatable property for alloy compositions.

    :group: utils
    """

    properties = []

    for p in fields(elementy.element.Element):
        if p.name in ["name", "symbol"]:
            continue
        if (
            p.type in [int, float, Optional[float]]
            and p.name not in properties
        ):
            properties.append(p.name)

    # Get all modules in metallurgy
    modules = inspect.getmembers(mg, inspect.ismodule)
    for module in modules:
        for func in list_functions(module[1]):
            argspec = inspect.getfullargspec(getattr(module[1], func))
            if (
                "alloy" in argspec.args
                and len(argspec.args) == 1
                and func not in properties
            ):
                properties.append(func)

    return properties


def calculate(
    alloy: Union[mg.Alloy, str, dict],
    property_name: str,
    uncertainty: bool = False,
) -> Union[float, None, List[Union[float, None]]]:
    """Returns the a particular property calculated for an alloy, using other
    calculation functions provided by metallurgy. The property_name must match a
    function name in a metallurgy module, or be suffixed with _linearmix or
    _deviation to denote linear mixture or deviation of an elemental property.

    :group: calculations

    Parameters
    ----------

    alloy
        The alloy for which to calculate a property.
    property_name
        The property to calculate for the alloy.
    uncertainty
        If using a cerebral model, activate dropout layers during inference and
        gather uncertainty information.
    """

    # If we have a predictive model available to use
    if mg.get_model() is not None:
        import cerebral as cb

        # Check if the property is an output of the model, if so use the model
        # and return the prediction
        model = mg.get_model()
        if property_name in [
            f["name"] for f in cb.models.get_model_prediction_features(model)
        ]:
            return cb.models.predict(model, alloy, uncertainty=uncertainty)[
                property_name
            ]

    # Check for simple linear mixture or deviations of elemental properties
    if "_linearmix" in property_name:
        return mg.linear_mixture(alloy, property_name.split("_linearmix")[0])
    if "_deviation" in property_name:
        return mg.deviation(alloy, property_name.split("_deviation")[0])

    # Otherwise, check all function names to find a match to the property
    property_function = get_property_function(property_name)
    if property_function is not None:
        return property_function(alloy)

    # If all else fails, try a simple linear mixture again
    return mg.linear_mixture(alloy, property_name)



import metallurgy as mg


def test_calculate_property_name():
    assert mg.calculate(
        {"Cu": 0.5, "Zr": 0.5}, "melting_temperature_linearmix"
    ) == mg.linear_mixture({"Cu": 0.5, "Zr": 0.5}, "melting_temperature")

    assert mg.calculate(
        {"Cu": 0.5, "Zr": 0.5}, "melting_temperature_deviation"
    ) == mg.deviation({"Cu": 0.5, "Zr": 0.5}, "melting_temperature")



"""``metallurgy.linear_mixture``
=============================

Module enabling calculation of linear mixtures of elemental properties, as an
approximation to the properties of an alloy.

"""

from typing import Union, List
from numbers import Number
from collections.abc import Iterable

import metallurgy as mg


def linear_mixture(
    alloy: Union[mg.Alloy, str, dict], property_name: str
) -> Union[Number, None, List[Union[Number, None]]]:
    """Returns the linear mixture of a particular elemental property of an
    alloy.

    See equation 5 of the paper "Machine-learning improves understanding of
    glass formation in metallic systems" for definition of the linear mixture:
    https://pubs.rsc.org/en/content/articlelanding/2022/dd/d2dd00026a

    :group: calculations

    Parameters
    ----------

    alloy : Alloy, str, dict
        The alloy for which to calculate the linear mixture.
    property_name : str
        The elemental property to calcualate the linear mixture of.

    """

    # If a list of alloys is given, return a list of linear mixture data
    if isinstance(alloy, Iterable) and not isinstance(alloy, (str, dict)):
        return [linear_mixture(a, property_name) for a in list(alloy)]

    # Convert input alloy to an Alloy instance if not already
    elif not isinstance(alloy, mg.Alloy):
        alloy = mg.Alloy(alloy)

    # Calculate the linear mixture
    mixed_property = 0.0
    for element in alloy.elements:
        # Get the current element's value of the property from elementy
        value = mg.periodic_table.elements[element][property_name]

        # Return None if no data for the element
        if value is None:
            return None
        # Use the first entry for the property if multiple
        elif isinstance(value, list):
            value = value[0]

        # If not numerical data, return None
        if not isinstance(value, Number):
            return None

        # Element's contribution to the linear mixture, weighted by composition
        mixed_property += alloy.composition[element] * value

    return mixed_property
